/**
 * Copyright (C) 2010 Andy Spencer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _XOPEN_SOURCE
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <libsoup/soup.h>
#include <glib/gstdio.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <json-glib/json-glib.h>

#ifndef UI
#define UI "/usr/local/share/twit/twit.ui"
#endif

/* URIs */
const gchar *post_link     = "http://twitter.com/%s/status/%lld";
const gchar *user_link     = "http://twitter.com/%s";

const gchar *user_api      = "https://api.twitter.com/1/statuses/user_timeline/%lld.json";
const gchar *list_api      = "https://api.twitter.com/1/%s/lists/%lld/statuses.json";

const gchar *ff_apis[][2] = {
	{"tl_friends_menu",   "https://api.twitter.com/1/statuses/friends.json"},
	{"tl_followers_menu", "https://api.twitter.com/1/statuses/followers.json"},
};

const gchar *timeline_apis[][2] = {
	{"tl_home",           "https://api.twitter.com/1/statuses/home_timeline.json"},
	{"tl_public",         "https://api.twitter.com/1/statuses/public_timeline.json"},
	{"tl_friends",        "https://api.twitter.com/1/statuses/friends_timeline.json"},
	{"tl_user",           "https://api.twitter.com/1/statuses/user_timeline.json"},
	{"tl_mentions",       "https://api.twitter.com/1/statuses/mentions.json"},
	{"tl_rt_by",          "https://api.twitter.com/1/statuses/retweeted_by_me.json"},
	{"tl_rt_to",          "https://api.twitter.com/1/statuses/retweeted_to_me.json"},
	{"tl_rt_of",          "https://api.twitter.com/1/statuses/retweets_of_me.json"},
	{"tl_mesg_recv",      "https://api.twitter.com/1/direct_messages.json"},
	{"tl_mesg_sent",      "https://api.twitter.com/1/direct_messages/sent.json"},
};

/* Constants */
struct {
	char *name;
	char *path;
	guint key;
	GdkModifierType mask;
} keymap[] = {
	{"refresh", "<Twitter>/refresh", GDK_r, GDK_CONTROL_MASK},
	{"clear",   "<Twitter>/clear",   GDK_l, GDK_CONTROL_MASK},
	{"quit",    "<Twitter>/quit",    GDK_q, 0},
};

/* Globals */
SoupSession *soup       = NULL;
GtkBuilder  *builder    = NULL;
GKeyFile    *prefs      = NULL;
gchar       *prefs_path = NULL;
const gchar *timeline   = NULL;
gint64       cleared    = 1;


/********
 * JSON *
 ********/
struct {
	const gchar *ptrn;
	const gchar *repl;
	GRegex *re;
} regexes[] = {
	{"&lt;",  "<", NULL},
	{"&gt;",  ">", NULL},
	{"&amp;", "&", NULL},
};
gchar *json_escape(const gchar *_str)
{
	if (!_str || !_str[0])
		return NULL;
	char *str = g_strdup(_str);
	for (int i = 0; i < G_N_ELEMENTS(regexes); i++) {
		if (!regexes[i].re)
			regexes[i].re = g_regex_new(regexes[i].ptrn, 0, 0, NULL);
		gchar *tmp = g_regex_replace(regexes[i].re, str, -1, 0, regexes[i].repl, 0, NULL);
		g_free(str);
		str = tmp;
	}
	gchar *rval = g_markup_escape_text(str, -1);
	g_free(str);
	return rval;
}
static JsonNode *json_getv(JsonNode *node, va_list ap)
{
	const gchar *memb = va_arg(ap, const gchar *);
	if (memb == NULL)
		return node;
	JsonObject *object = json_node_get_object(node);
	if (!object) return NULL;
	JsonNode   *child  = json_object_get_member(object, memb);
	if (!child)  return NULL;
	return json_getv(child, ap);
}
static JsonNode *json_get(JsonNode *node, ...)
{
	va_list ap;
	va_start(ap, node);
	node = json_getv(node, ap);
	va_end(ap);
	return node;
}
#define json_get_str(...) ({ \
	JsonNode *node = json_get(__VA_ARGS__); \
	node ? json_node_get_string(node) : NULL; \
})
#define json_get_strdup(...) ({ \
	JsonNode *node = json_get(__VA_ARGS__); \
	node ? g_strdup(json_node_get_string(node)) : NULL; \
})
#define json_get_escaped(...) ({ \
	JsonNode *node = json_get(__VA_ARGS__); \
	node ? json_escape(json_node_get_string(node)) : NULL; \
})
#define json_get_int(...) ({ \
	JsonNode *node = json_get(__VA_ARGS__); \
	node ? json_node_get_int(node) : 0; \
})


/***********
 * Helpers *
 ***********/
#define g_signal_connect_free(inst, sig, cb, data) \
	g_signal_connect_data(inst, sig, cb, data, (GClosureNotify)g_free, 0)
gboolean refresh(gpointer);
void save_prefs(GKeyFile *prefs, const gchar *path)
{
	gsize length;
	gchar *data = g_key_file_to_data(prefs, &length, NULL);
	g_file_set_contents(path, data, length, NULL);
}
void set_timeline(GtkWidget *widget, gpointer _uri)
{
	if (!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(widget)))
		return;

	/* Reset globals */
	cleared  = 1;
	timeline = _uri;

	/* Save timeline and reset cleared time */
	g_key_file_set_double(prefs, "refresh", "cleared",  cleared);
	g_key_file_set_string(prefs, "refresh", "timeline", timeline);
	save_prefs(prefs, prefs_path);

	/* Load timeline */
	g_message("Loading '%s'", timeline);
	refresh(NULL);
}
void show_uri(GtkWidget *widget, gpointer _uri)
{
	/* Because gtk_show_uri is broken */
	gchar *uri = g_shell_quote(_uri);
	g_message("Showing %s", uri);
	gchar *cmd = g_strdup_printf("xdg-open %s", uri);
	g_spawn_command_line_async(cmd, NULL);
	g_free(uri);
	g_free(cmd);

	//GError *err = NULL;
	//if (!gtk_show_uri(NULL, uri, GDK_CURRENT_TIME, &err)) {
	//	g_warning("error showing `%s' %s", uri, err->message);
	//	g_error_free(err);
	//}
}


/****************
 * Post buttons *
 ****************/
void icon_cache_cb(SoupSession *session, SoupMessage *msg, gpointer icon)
{
	SoupBuffer *buf = soup_message_body_flatten(msg->response_body);
	SoupURI    *uri = soup_message_get_uri(msg);
	gchar *name, tmpl[64];
	g_snprintf(tmpl, sizeof(tmpl), "XXXXXX%s", g_strrstr(uri->path, "."));
	gint fd = g_file_open_tmp(tmpl, &name, NULL);
	if (fd > 0) {
		write(fd, buf->data, buf->length);
		close(fd);
		GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(name, NULL);
		g_object_set_data(G_OBJECT(icon), "pixbuf", pixbuf);
		gtk_tooltip_trigger_tooltip_query(gdk_display_get_default());
		g_unlink(name);
	}
	g_free(name);
}
gboolean make_button_tooltip(GtkStatusIcon *icon, gint x, gint y, gboolean kb,
		GtkTooltip *tip, JsonNode *post)
{
	gchar *name = json_get_escaped(post, "user", "name", NULL);
	gchar *web  = json_get_escaped(post, "user", "url", NULL);
	gchar *bio  = json_get_escaped(post, "user", "description", NULL);
	GString *markup = g_string_new("");
	if (name) g_string_append_printf(markup, "<b>Name</b> %s\n", name);
	if (web)  g_string_append_printf(markup, "<b>Web</b> %s\n",  web);
	if (bio)  g_string_append_printf(markup, "<b>Bio</b> %s\n",  bio);
	gtk_tooltip_set_markup(tip, markup->str);
	g_string_free(markup, TRUE);

	GdkPixbuf *pixbuf = g_object_get_data(G_OBJECT(icon), "pixbuf");
	if (pixbuf) {
		gtk_tooltip_set_icon(tip, pixbuf);
	} else {
		gtk_tooltip_set_icon_from_stock(tip, GTK_STOCK_INFO, GTK_ICON_SIZE_DIALOG);
		if (!g_object_get_data(G_OBJECT(icon), "loading")) {
			const gchar *uri = json_get_str(post, "user", "profile_image_url", NULL);
			SoupMessage *msg = soup_message_new(SOUP_METHOD_GET, uri);
			soup_session_queue_message(soup, msg, icon_cache_cb, icon);
			g_object_set_data(G_OBJECT(icon), "loading", NULL+1);
		}
	}

	return TRUE;
}
GtkWidget *make_button_menu(JsonNode *post)
{
	gchar *timeline_uri = g_strdup_printf(user_api,
			json_get_int(post, "user", "id", NULL));
	gchar *user_uri = g_strdup_printf(user_link,
			json_get_str(post, "user", "screen_name", NULL));
	gchar *post_uri = g_strdup_printf(post_link,
			json_get_str(post, "user", "screen_name",  NULL),
			json_get_int(post, "id",  NULL));
	gchar *web_uri = json_get_strdup(post, "user", "url", NULL);

	GtkWidget *menu = gtk_menu_new();
	GtkWidget *timeline_item = gtk_radio_menu_item_new_with_label_from_widget(
			GTK_RADIO_MENU_ITEM(gtk_builder_get_object(builder, "tl_home")),
			"Timeline");
	GtkWidget *user_item = gtk_menu_item_new_with_label("View user");
	GtkWidget *post_item = gtk_menu_item_new_with_label("View post");
	g_signal_connect_free(timeline_item, "activate", G_CALLBACK(set_timeline), timeline_uri);
	g_signal_connect_free(user_item,     "activate", G_CALLBACK(show_uri),     user_uri);
	g_signal_connect_free(post_item,     "activate", G_CALLBACK(show_uri),     post_uri);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), timeline_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), user_item);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), post_item);
	if (web_uri) {
		GtkWidget *web_item = gtk_menu_item_new_with_label("View web");;
		g_signal_connect_free(web_item, "activate", G_CALLBACK(show_uri), web_uri);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), web_item);
	}
	return menu;
}
gboolean make_button_press(GtkWidget *button, GdkEventButton *event, gpointer post)
{
	if (event->button == 3 && event->type == GDK_BUTTON_PRESS) {
		GtkWidget *menu = make_button_menu(post);
		gtk_menu_attach_to_widget(GTK_MENU(menu), button, NULL);
		gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
				event->button, event->time);
		gtk_widget_show_all(menu);
	}
	return FALSE;
}
GtkWidget *make_button(JsonNode *post)
{
	/* Filter */
	static gboolean loaded = FALSE;
	static GRegex  *filter = NULL;
	if (!loaded) {
		gchar *filter_str = g_key_file_get_string(prefs, "general", "filter", NULL);
		if (filter_str)
			filter = g_regex_new(filter_str,  0, 0, NULL);
		loaded = TRUE;
	}
	;
	const gchar *screenname = json_get_str(post, "user", "screen_name", NULL);
	if (filter && screenname && !g_regex_match(filter, screenname, 0, NULL))
		return NULL;

	/* Dup node */
	post = json_node_copy(post);

	/* Pretty print */
	struct tm tm;
	strptime(json_get_str(post, "created_at", NULL),
			"%a %b %d %H:%M:%S %z %Y", &tm);
	time_t age = time(NULL) - mktime(&tm);
	gchar *time_str = g_strdup_printf("-%ld:%02ld:%02ld", age/(60*60), (age/60)%60, age%60);

	gchar *name_str = json_get_escaped(post, "user", "name", NULL);
	if (!name_str)
		name_str = g_strdup_printf("%s → %s",
				json_get_str(post, "sender",    "name", NULL),
				json_get_str(post, "recipient", "name", NULL));

	gchar *text_str = json_get_escaped(post, "text", NULL);

	/* Create button */
	gchar *post_uri = g_strdup_printf(post_link, screenname,
			json_get_int(post, "id",  NULL));
	GtkWidget *btn = gtk_button_new();
	g_object_set_data_full(G_OBJECT(btn), "post",
			post, (GDestroyNotify)json_node_free);
	g_signal_connect_free(btn, "clicked", G_CALLBACK(show_uri), post_uri);
	g_signal_connect_after(btn, "button-press-event", G_CALLBACK(make_button_press), post);
	g_object_set(btn, "has-tooltip", TRUE, NULL);
	g_signal_connect(btn, "query-tooltip", G_CALLBACK(make_button_tooltip), post);

	/* Add label */
	gchar *markup = g_strdup_printf(
			"<span font_weight=\"bold\" font=\"Mono 8\">(%s)</span> "
			"<span font_weight=\"bold\" font=\"Sans 10\">%s</span>: "
			"%s", time_str, name_str, text_str);
	GtkWidget *label = gtk_label_new(markup);
	gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
	gtk_label_set_selectable(GTK_LABEL(label), TRUE);
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	gtk_label_set_line_wrap_mode(GTK_LABEL(label), PANGO_WRAP_WORD);
	gtk_container_add(GTK_CONTAINER(btn), label);

	g_free(time_str);
	g_free(name_str);
	g_free(text_str);
	g_free(markup);
	return btn;
}


/************
 * Timeline *
 ************/
void size_allocate(GtkWidget *widget, GtkAllocation *alloc, gpointer user_data)
{
	GList *kids = gtk_container_get_children(GTK_CONTAINER(widget));
	for (GList *cur = kids; cur; cur = cur->next)
		gtk_widget_set_size_request(GTK_WIDGET(cur->data), alloc->width-4, -1);
	g_list_free(kids);
}
void refresh_add(JsonArray *_, guint index, JsonNode *post, gpointer _tweets)
{
	GtkWidget *tweets = _tweets;
	GtkWidget *btn = make_button(post);
	if (!btn) return;
	gtk_box_pack_start(GTK_BOX(tweets), btn, FALSE, FALSE, 0);
	gtk_box_reorder_child(GTK_BOX(tweets), btn, -1);
	gtk_widget_set_size_request(GTK_WIDGET(tweets), 0, -1);
	g_signal_connect(btn, "size-allocate", G_CALLBACK(size_allocate),   NULL);
	gtk_widget_show_all(btn);
}
void refresh_cb(SoupSession *session, SoupMessage *msg, gpointer _)
{
	//g_message("refresh_cb");
	if (msg->status_code != 200) {
		/* Clear statusbar */
		GObject *status = gtk_builder_get_object(builder, "status");
		gtk_statusbar_pop(GTK_STATUSBAR(status), 0);
		gtk_statusbar_push(GTK_STATUSBAR(status), 0, "Error loading");
		return;
	}

	GObject *tweets = gtk_builder_get_object(builder, "tweets");

	/* Clear old posts */
	GList *old = gtk_container_get_children(GTK_CONTAINER(tweets));
	for (GList *cur = old; cur; cur = cur->next)
		gtk_container_remove(GTK_CONTAINER(tweets), cur->data);
	g_list_free(old);

	/* Parse Json */
	SoupBuffer *buf = soup_message_body_flatten(msg->response_body);
	JsonParser *parser = json_parser_new();
	json_parser_load_from_data(parser, buf->data, buf->length, NULL);
	JsonNode  *root  = json_parser_get_root(parser);
	JsonArray *posts = json_node_get_array(root);
	json_array_foreach_element(posts, refresh_add, tweets);
	g_object_unref(parser);
	soup_buffer_free(buf);

	/* Clear statusbar */
	GObject *status = gtk_builder_get_object(builder, "status");
	gtk_statusbar_pop(GTK_STATUSBAR(status), 0);
	gtk_widget_hide(GTK_WIDGET(status));
}
gboolean refresh(gpointer user_data)
{
	/* Add status message */
	GObject *status = gtk_builder_get_object(builder, "status");
	gtk_statusbar_pop(GTK_STATUSBAR(status), 0); // Clear errors
	gtk_statusbar_push(GTK_STATUSBAR(status), 0, "Loading..");
	gtk_widget_show(GTK_WIDGET(status));

	/* Queue download */
	gchar *uri = g_strdup_printf("%s?since_id=%lld", timeline, cleared);
	SoupMessage *msg = soup_message_new(SOUP_METHOD_GET, uri);
	soup_session_queue_message(soup, msg, refresh_cb, NULL);
	g_free(uri);

	/* Queue next refresh */
	static guint tag = 0;
	gint update = g_key_file_get_integer(prefs, "general", "update", NULL);
	if (update == 0) update = 60*3;
	if (tag) g_source_remove(tag);
	tag = g_timeout_add_seconds(update, refresh, NULL);

	return FALSE;
}
gboolean clear(gpointer user_data)
{
	GObject *tweets = gtk_builder_get_object(builder, "tweets");
	GList *old = gtk_container_get_children(GTK_CONTAINER(tweets));
	for (GList *cur = old; cur; cur = cur->next) {
		JsonNode *post = g_object_get_data(G_OBJECT(cur->data), "post");
		cleared = MAX(cleared, json_get_int(post, "id", NULL));
		gtk_container_remove(GTK_CONTAINER(tweets), cur->data);
	}
	g_list_free(old);

	/* Clear time */
	g_key_file_set_double(prefs, "refresh", "cleared",  cleared);
	save_prefs(prefs, prefs_path);

	refresh(NULL);
	return FALSE;
}


/***********
 * Loading *
 ***********/
void load_user(JsonArray *_, guint index, JsonNode *user, gpointer _menu)
{
	GtkMenuShell *menu = _menu;
	GtkWidget *child = gtk_radio_menu_item_new_with_label_from_widget(
			GTK_RADIO_MENU_ITEM(gtk_builder_get_object(builder, "tl_home")),
			json_get_str(user, "name", NULL));
	gchar *uri = g_strdup_printf(user_api, json_get_int(user, "id",  NULL));
	g_signal_connect_free(child, "activate", G_CALLBACK(set_timeline), uri);
	gtk_menu_shell_append(menu, child);
	gtk_widget_show_all(child);
}
void load_menu(SoupSession *session, SoupMessage *msg, gpointer menu)
{
	if (msg->status_code != 200) {
		g_warning("Error loading friends/followers");
		return;
	}
	SoupBuffer *buf = soup_message_body_flatten(msg->response_body);
	JsonParser *parser = json_parser_new();
	json_parser_load_from_data(parser, buf->data, buf->length, NULL);
	JsonNode  *root  = json_parser_get_root(parser);
	JsonArray *users = json_node_get_array(root);
	json_array_foreach_element(users, load_user, menu);
	soup_buffer_free(buf);
	g_object_unref(parser);
}
void auth(SoupSession *session, SoupMessage *msg, SoupAuth *auth,
		gboolean retrying, gpointer user_data)
{
	GStaticMutex mutex = G_STATIC_MUTEX_INIT;
	if (!g_static_mutex_trylock(&mutex))
		return;

	GObject *dialog = gtk_builder_get_object(builder, "auth_win");
	GObject *user   = gtk_builder_get_object(builder, "auth_user");
	GObject *pass   = gtk_builder_get_object(builder, "auth_pass");
	GObject *save   = gtk_builder_get_object(builder, "auth_save");

	/* Try stored defaults */
	if (!retrying) {
		const gchar *user_str = gtk_entry_get_text(GTK_ENTRY(user));
		const gchar *pass_str = gtk_entry_get_text(GTK_ENTRY(pass));
		soup_auth_authenticate(auth, user_str, pass_str);
		return;
	}

	/* Prompt for a password */
	soup_session_pause_message(session, msg);
	switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
	case GTK_RESPONSE_OK: {
		const gchar *user_str = gtk_entry_get_text(GTK_ENTRY(user));
		const gchar *pass_str = gtk_entry_get_text(GTK_ENTRY(pass));
		soup_auth_authenticate(auth, user_str, pass_str);
		if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(save))) {
			g_key_file_set_string(prefs, "auth", "user", user_str);
			g_key_file_set_string(prefs, "auth", "pass", pass_str);
			save_prefs(prefs, prefs_path);
		}
		break;
	}
	default:
		soup_session_abort(session);
		break;
	}
	soup_session_unpause_message(session, msg);
	gtk_widget_hide(GTK_WIDGET(dialog));

	g_static_mutex_unlock(&mutex);
}


/********
 * Main *
 ********/
gint main(gint argc, gchar **argv)
{
	gtk_init(&argc, &argv);
	g_thread_init(NULL);

	/* Load soup */
	soup = soup_session_async_new();
	g_object_set(soup, "timeout", 10, NULL);
	g_signal_connect(soup, "authenticate", G_CALLBACK(auth), NULL);

	/* Load UI */
	builder = gtk_builder_new();
	gtk_builder_add_from_file(builder, UI, NULL);
	gtk_builder_connect_signals(builder, NULL);
	GObject *win = gtk_builder_get_object(builder, "win");

	/* Connect accelerators.. wow, this is ugly */
	GtkAccelGroup *group = gtk_accel_group_new();
	for (int i = 0; i < G_N_ELEMENTS(keymap); i++) {
		GObject *action = gtk_builder_get_object(builder, keymap[i].name);
		gtk_accel_map_add_entry(keymap[i].path, keymap[i].key, keymap[i].mask);
		gtk_action_set_accel_path(GTK_ACTION(action), keymap[i].path);
		gtk_action_set_accel_group(GTK_ACTION(action), group);
		gtk_action_connect_accelerator(GTK_ACTION(action));
	}
	gtk_window_add_accel_group(GTK_WINDOW(win), group);

	/* Load prefs */
	prefs = g_key_file_new();
	gchar *dir = g_build_filename(g_get_user_config_dir(), "twit", NULL);
	prefs_path = g_build_filename(dir, "config.ini", NULL);
	g_mkdir_with_parents(dir, 0755);
	g_key_file_load_from_file(prefs, prefs_path, 0, NULL);
	g_free(dir);
	cleared = g_key_file_get_double(prefs, "refresh", "cleared", NULL);
	if (cleared == 0) cleared = 1;
	timeline = g_key_file_get_string(prefs, "refresh", "timeline", NULL);
	if (!timeline) timeline = timeline_apis[0][1];

	GObject *user = gtk_builder_get_object(builder, "auth_user");
	GObject *pass = gtk_builder_get_object(builder, "auth_pass");
	gchar *user_str = g_key_file_get_string(prefs, "auth", "user", NULL);
	gchar *pass_str = g_key_file_get_string(prefs, "auth", "pass", NULL);
	if (user_str) gtk_entry_set_text(GTK_ENTRY(user), user_str);
	if (pass_str) gtk_entry_set_text(GTK_ENTRY(pass), pass_str);
	g_free(user_str);
	g_free(pass_str);

	/* Finish setup */
	for (int i = 0; i < G_N_ELEMENTS(timeline_apis); i++) {
		const gchar **uris = timeline_apis[i];
		GObject *item = gtk_builder_get_object(builder, uris[0]);
		g_signal_connect(item, "activate",
			G_CALLBACK(set_timeline), (gpointer)uris[1]);
	}
	for (int i = 0; i < G_N_ELEMENTS(ff_apis); i++) {
		const gchar **uris = ff_apis[i];
		GObject *item = gtk_builder_get_object(builder, uris[0]);
		GtkWidget *menu = gtk_menu_new();
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
		SoupMessage *msg = soup_message_new(SOUP_METHOD_GET, uris[1]);
		soup_session_queue_message(soup, msg, load_menu, menu);
	}

	/* Go! */
	g_idle_add(refresh, NULL);
	gtk_widget_show_all(GTK_WIDGET(win));
	gtk_main();

	return 0;
}
